{
 uNotifer.pas

  This source code is Copyright (C) 2004,  Kyley Harris

  THIS SOFTWARE IS PROVIDED 'AS-IS', WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTY.  IN NO EVENT WILL THE AUTHOR BE HELD LIABLE FOR ANY DAMAGES
  ARISING FROM THE USE OF THIS SOFTWARE.

  PERMISSION IS GRANTED TO ANYONE WHO HAS RECEIVED PERMISSION FROM THE AUTHOR
  IN WRITING, OR PURCHASED A LICENCE TO USE FROM THE AUTHOR TO USE THIS
  SOFTWARE FOR ANY PURPOSE, INCLUDING COMMERCIAL APPLICATIONS, AND TO ALTER IT
  FREELY, SUBJECT TO THE FOLLOWING RESTRICTIONS:

  THE ORIGIN OF THIS SOFTWARE MUST NOT BE MISREPRESENTED; YOU MUST NOT
     CLAIM THAT YOU WROTE THE ORIGINAL SOFTWARE. IF YOU USE THIS SOFTWARE
     IN A PRODUCT, AN ACKNOWLEDGMENT IN THE PRODUCT DOCUMENTATION WOULD BE
     APPRECIATED BUT IS NOT REQUIRED.
  ALTERED SOURCE VERSIONS MUST BE PLAINLY MARKED AS SUCH, AND MUST NOT BE
     MISREPRESENTED AS BEING THE ORIGINAL SOFTWARE.

  ANY CHANGES TO THE CODE WOULD BE APPRECIATED TO BE EMAILED TO THE AUTHOR
  FOR CONSIDERATION TO ANY IMPROVEMENTS.

  Kyley Harris, Auckland, New Zealand
  Kyley@HarrisSoftware.com

################################################################################

 The TNotifier class provides an easy to use method of allowing many objects
 to listen to an event from a source object.

 TNotifier does not need to be instantiated. Just use the class functions.


}
unit uNotifier;

interface

uses
  {$IFDEF HSS}
  uItem,
  {$ENDIF}
  SysUtils, Classes, SyncObjs;
type
  TMessageClass = class of TMessage;

  TMessage = class(TObject)
  protected
    class function RegisterMessage : Integer;
    class function UnregisterMessage : Integer;
  public
    class function TextToMessage(AValue : String) : TMessageClass;
    class function AsString : String; virtual; abstract;
  end;

  TUpdateMessage = class(TMessage)
  public
    class function AsString : String; override;
  end;

  {@@ This is the event type for receiving notifications. Broadcaster is the source,
    AMessage is a user defined string, and Data is an optional attached object message.
    Data should be treated as if the lifetime is only for this method. Information should
    be copied from it, not referenced. }
  {$IFDEF HSS}
  TNotificationEvent = procedure (Broadcaster:TObject;AMessage:TMessageClass;Data:TItem) of object;
  {$ELSE}
  TNotificationEvent = procedure (Broadcaster:TObject;AMessage:TMessageClass;Data:TObject) of object;
  {$ENDIF}


  {@@ IMPORTANT.... Any user of AddClassListener and AddListener must reverse
  the call by calling RemClassListener, or RemListener before freeing the instance
  of the listener, otherwise all types of crap will happen }
  TNotifier = class(TObject)
  public
    {@@ use these to monitor changes to a specific instance of a class.
      ABroadcaster is the instance that will be sending notifications. This can be any subclass
      of TObject. ABroadcaster does not have to be registered in any way.
      Life management is handled via the AddListener, RemListener calls.

      BindMethod attaches an object as a listener via the bind method event. All events
      which have ABroadcaster as the source will be sent to this event.
    }
    class procedure AddListener(ABroadcaster:TObject;BindMethod:TNotificationEvent);
    {@@ remove a previously bound event from the broadcaster. }
    class procedure RemListener(ABroadcaster:TObject;BindMethod:TNotificationEvent);

    {@@ use these to monitor changes on any instance of the specified class. Binding the
      event to this method will trigger BindMethod any time BroadCast is called with
      a broadcaster that Matches AClass }
    class procedure AddClassListener(AClass:TClass;BindMethod:TNotificationEvent);
    class procedure RemClassListener(AClass:TClass;BindMethod:TNotificationEvent);

    {@@ Use this to trigger events to any BindEvent attached to the memory pointer
      reference of ABroadcaster. It is very important to Dereference Listeners
      correctly, otherwise messages could be sent to the wrong events depending
      on memory pointer allocations.}
  {$IFDEF HSS}
    class procedure BroadCast(ABroadCaster:TObject;AMessage:TMessageClass;Data:TItem);
  {$ELSE}
    class procedure BroadCast(ABroadCaster:TObject;AMessage:TMessageClass;Data:TObject);
  {$ENDIF}  
  end;

implementation

var
  _CS:TCriticalSection;

function CS:TCriticalSection;
begin
  if _CS = nil then
    _CS := TCriticalSection.Create;
  result := _CS;
end;

{ TNotifier }

type
  TListenerItem = class(TCollectionItem)
  private
    FBindMethod: TNotificationEvent;
    FListener: TObject;
  published
    property BindMethod:TNotificationEvent read FBindMethod;
    property AListener:TObject read FListener;
  end;

  TListeners = class(TCollection)
  private
    function GetItem(Index: Integer): TListenerItem;
    procedure SetItem(Index: Integer; const Value: TListenerItem);
  public
    constructor Create;
    function Add: TListenerItem;
    procedure Remove(AListener:TObject;ABinder:TNotificationEvent);
    property Items[Index: Integer]: TListenerItem read GetItem write SetItem;default;

  end;


  TBroadCastList = class(TStringList)
  private
    function GetListeners(i: integer): TListeners;
  public
    property Listeners[i:integer]:TListeners read GetListeners;
  end;

var
  FBroadCastList:TBroadCastList;
  FClassBroadCastList:TBroadCastList;
  SingletonMessageList : TStringList;

function MessageList : TStringList;
begin
  if (SingletonMessageList = nil) then
  begin
    SingletonMessageList := TStringList.Create;
    SingletonMessageList.Sorted := True;
    SingletonMessageList.Duplicates := dupIgnore;
  end;

  Result := SingletonMessageList;
end;

function ObjHex(AObj:TObject):string;
begin
  result := IntToHex(integer(AOBJ),8);
end;

function BroadCastList:TBroadCastList;
begin
  if FBroadCastList = nil then
  begin
    FBroadCastList := TBroadCastList.Create;
    FBroadCastList.Sorted := true;

  end;
  result := FBroadCastList;
end;

function ClassBroadCastList:TBroadCastList;
begin
  if FClassBroadCastList = nil then
  begin
    FClassBroadCastList := TBroadCastList.Create;
    FClassBroadCastList.Sorted := true;
  end;
  result := FClassBroadCastList;
end;

function AddBroadcaster(ABroadcaster: TObject):TListeners;
var
  i:integer;
begin
  with BroadCastList do
  begin
    i := IndexOf(ObjHex(ABroadcaster));
    if i = -1 then
    begin
      result := TListeners.Create;
      AddObject(ObjHex(ABroadcaster),result);
    end else
      result := Listeners[i];
  end;
end;

function AddClassBroadcaster(AClass: TClass):TListeners;
var
  i:integer;
begin
  with ClassBroadCastList do
  begin
    i := IndexOf(AClass.ClassName);
    if i = -1 then
    begin
      result := TListeners.Create;
      AddObject(AClass.ClassName,result);
    end else
      result := Listeners[i];
  end;
end;

class procedure TNotifier.AddClassListener(AClass: TClass; BindMethod: TNotificationEvent);
var
  Listen:Tlisteners;
  AListen:TListenerItem;
  AListener:TObject;
begin
  AListener := TMethod(BindMethod).Data;
  CS.Acquire;
  try
    Listen := AddClassBroadcaster(AClass);
    AListen := Listen.Add;
    AListen.FBindMethod := BindMethod;
    AListen.FListener := AListener;
  finally
    CS.Release;
  end;


end;

class procedure TNotifier.AddListener(ABroadcaster:TObject; BindMethod: TNotificationEvent);
var
  Listen:Tlisteners;
  AListen:TListenerItem;
  AListener:TObject;
begin
  AListener := TMethod(BindMethod).Data;
  CS.Acquire;
  try
    Listen := AddBroadcaster(ABroadcaster);
    AListen := Listen.Add;
    AListen.FBindMethod := BindMethod;
    AListen.FListener := AListener;
  finally
    CS.Release;
  end;


end;

{$IFDEF HSS}
class procedure TNotifier.BroadCast(ABroadCaster:TObject;AMessage:TMessageClass; Data: TItem);
{$ELSE}
class procedure TNotifier.BroadCast(ABroadCaster:TObject;AMessage:TMessageClass; Data: TObject);
{$ENDIF}

var
  i:integer;
  l:TListeners;
  AClass:TClass;
begin

  CS.Acquire;
  try
    i := BroadCastList.IndexOf(ObjHex(ABroadCaster));
    if i <> -1 then
    begin
      l := BroadcastList.Listeners[i];
      for i := l.Count -1 downto 0 do
        l.Items[i].BindMethod(ABroadCaster,AMessage,Data);

    end;

    if Assigned(ABroadcaster) then
    begin
      AClass := ABroadCaster.ClassType;

      repeat
        i := ClassBroadCastList.IndexOf(AClass.ClassName);
        if i <> -1 then
        begin
          l := ClassBroadcastList.Listeners[i];
          for i := l.Count -1 downto 0 do
            l.Items[i].BindMethod(ABroadCaster,AMessage,Data);

        end;
        AClass := AClass.ClassParent;
      until AClass = TObject;
    end;

  finally
    CS.Release;
  end;


end;

class procedure TNotifier.RemClassListener(AClass: TClass; BindMethod: TNotificationEvent);
var
  i:integer;
  AListener:TObject;
begin
  AListener := TMethod(BindMethod).Data;
  CS.Acquire;
  try
    i := ClassBroadCastList.IndexOf(AClass.ClassName);
    if i <> -1 then
    begin
      ClassBroadCastList.Listeners[i].Remove(AListener,BindMethod);
      if ClassBroadCastList.Listeners[i].Count = 0 then
      begin
        ClassBroadCastList.Objects[i].Free;
        ClassBroadCastList.Delete(i);
      end;
    end;
  finally

    CS.Release;
  end;

end;


class procedure TNotifier.RemListener(ABroadcaster: TObject; BindMethod: TNotificationEvent);
var
  i:integer;
  AListener:TObject;
begin
  AListener := TMethod(BindMethod).Data;
  CS.Acquire;
  try
    i := BroadCastList.IndexOf(ObjHex(ABroadcaster));
    if i <> -1 then
    begin
      BroadCastList.Listeners[i].Remove(AListener,BindMethod);
      if BroadCastList.Listeners[i].Count = 0 then
      begin
        BroadCastList.Objects[i].free;
        BroadCastList.Delete(i);
      end;
    end;
  finally

    CS.Release;
  end;

end;




{ TBroadCastList }

function TBroadCastList.GetListeners(i: integer): TListeners;
begin
  result := objects[i] as TListeners;
end;


{ TListeners }

function TListeners.Add: TListenerItem;
begin
  result := inherited add as TListenerItem;
end;

constructor TListeners.Create;
begin
  inherited Create(TListenerItem);
end;

function TListeners.GetItem(Index: Integer): TListenerItem;
begin
  result := inherited items[index] as TListenerItem;
end;

procedure TListeners.Remove(AListener: TObject;
  ABinder: TNotificationEvent);
var
  i:integer;

  function MethodEquals(AMethod,BMethod:TMethod):boolean;
  begin
    result := (AMethod.Code = BMethod.Code) and (AMethod.Data = BMethod.Data);
  end;
begin
  for i := 0 to Count -1 do
  begin
    if (MethodEquals(TMethod(Items[i].FBindMethod) ,TMethod(ABinder))) and (Items[i].FListener = AListener) then
    begin
      Items[i].Free;
      exit;
    end;

  end;
end;

procedure TListeners.SetItem(Index: Integer; const Value: TListenerItem);
begin
  inherited Items[index] := Value;
end;

{ TMessage }

class function TMessage.RegisterMessage: Integer;
begin
  CS.Enter;
  try
    Result := MessageList.AddObject(self.AsString,TObject(self));
  finally
    CS.Leave;
  end;
end;

class function TMessage.TextToMessage(AValue: String): TMessageClass;
var
  MessageIndex : Integer;
begin
  Result := nil;
  CS.Enter;
  try
    MessageIndex := MessageList.IndexOf(AValue);
    if (MessageIndex <> -1) then
    begin
      Result := TMessageClass(MessageList.Objects[MessageIndex]);
    end;
  finally
    CS.Leave;
  end;
end;

class function TMessage.UnregisterMessage: Integer;
begin
  CS.Enter;
  try
    Result := MessageList.IndexOf(self.AsString);
    MessageList.Delete(Result);
  finally
    CS.Leave;
  end;
end;

{ TUpdateMessage }

class function TUpdateMessage.AsString: String;
begin
  Result := 'MEMO_UPDATE';
end;

initialization
  TUpdateMessage.RegisterMessage;

finalization
  TUpdateMessage.UnregisterMessage;
  SingletonMessageList.Free;

  if Assigned(FBroadCastList) then
  while FBroadCastList.Count > 0 do
  begin
    FBroadCastList.Objects[0].Free;
    FBroadCastList.Delete(0);
  end;
  FreeAndNil(FBroadCastList);
  if Assigned(FClassBroadCastList) then
  while FClassBroadCastList.Count > 0 do
  begin
    FClassBroadCastList.Objects[0].Free;
    FClassBroadCastList.Delete(0);
  end;
  FreeAndNil(FClassBroadCastList);
  FreeAndNil(_CS);
end.

